#include <oxstd.h>
#include "mcmc_sv.h"       // header for sv.dll
#include "arms.h"	       // header for arms.dll
#include "../ssfpack/ssfpack.h"  // header for ssfpack.dll
#include "oxprob.h"

decl rep = 30;
decl d_post;
decl k_post;

// Class for posterior of d	======================================================
class postd
{
	decl psi, Ainv, Qs, k, T, q;
	post_d(const d);
	set_other(const new_psi, const new_Ainv, const new_Qs, const new_k, const new_T, const new_q);
};

postd::post_d(const d)
{
	decl Cinv_d = zeros(q,q);
	decl prod_d = new array[T];
		 prod_d[0:T-1] = zeros(q,q);
	  
	decl Q_sqrt, Qu, Qd, Qv;
		for(decl t=0; t<T; ++t)
		{
			decsvd(Qs[t][][], &Qu, &Qd, &Qv);
			Q_sqrt = Qu * diag(Qd.^(d/2)) * Qv';
 
			prod_d[t][][] = Q_sqrt * invert(Qs[t+1][][]) * Q_sqrt;
			Cinv_d += prod_d[t][][];
		}
	 
	return psi*d - 0.5*k*trace(Ainv*Cinv_d);
}

postd::set_other(const new_psi, const new_Ainv, const new_Qs, const new_k, const new_T, const new_q)
{
	psi  = new_psi;
	Ainv = new_Ainv;
	Qs   = new_Qs;
	k    = new_k;
	T    = new_T;
	q    = new_q;
}

post_d(const d)
{
	return d_post -> post_d(d);
}
//================================================================================

// Class for posterior of k ======================================================
class postk
{
	decl lambda0, A, Cinv_k, sum_log_Q, T, q;
	post_k(const k);
	set_other_k(const new_lambda0, const new_A, const new_Cinv_k, const new_sum_log_Q, const new_T, const new_q); 
};

postk::post_k(const k)
{
	return -lambda0*k + 0.5*T*k*q*log(k/2) - 0.5*T*k*log(determinant(A))
		   - T*sumr(log(gammafact((k+1-range(1,q))/2))) + (k/2)*sum_log_Q - 0.5*k*trace(invert(A)*Cinv_k); 
}

postk::set_other_k(const new_lambda0, const new_A, const new_Cinv_k, const new_sum_log_Q, const new_T, const new_q)
{
	lambda0 = new_lambda0;
	A         = new_A;
	Cinv_k    = new_Cinv_k;
	sum_log_Q = new_sum_log_Q;
	T         = new_T;
	q         = new_q;
}

post_k(const k)
{
	return k_post -> post_k(k);
}
//================================================================================

// Wishart generator =============================================================
rwish(const df, const Sc)
{
	decl p = sizer(Sc);
	decl B = 0 * Sc;
 
	if(p>1)
	{
		decl i,j;
		for(j=0; j<p; ++j)
		{
			B[j][j] = sqrt( rangamma(1, 1, (df-j)/2, 1/2) );   
		}

		for(i=1; i<p; ++i)
		{
			for(j=0; j<i; ++j)
			{
				B[i][j] = rann(1,1);
			}
		}   
	}
 
	decl A = choleski(Sc);
 
	return A * B * B' * A';
}
//================================================================================

// Set initial values for Qt =====================================================
Initials(const T, const A, const k, const d)
{
	decl q = sizer(A);

	// Storages	========================
	decl initial     = new array[T+1];
	initial[0:T]   = zeros(q,q);
	//==================================

	decl Q = unit(q);
	decl Qu, Qd, Qv;
		 decsvd(Q, &Qu, &Qd, &Qv);

	decl Q_neg_sqrt = Qu*diag(Qd.^(-d/2))*Qv';
	decl S = (1/k) * Q_neg_sqrt * A * Q_neg_sqrt;
		 initial[0] = Q;
	  
	decl Su, Sd, Sv, Qinv;
	for(decl i=1; i<=T; ++i)
	{
		Qinv = rwish(k, S);   
		Q = invert(Qinv);
		decsvd(Q, &Qu, &Qd, &Qv);
	   
		Q_neg_sqrt = Qu*diag(Qd.^(-d/2))*Qv';
		S = (1/k) * Q_neg_sqrt * A * Q_neg_sqrt;

		initial[i] = Q;
	}
	return initial;
}
//================================================================================

main()
{
	decl n_samples = 30000;
	decl burn = 10000;

	//Observations
	decl z = loadmat("C:\Research\DHKT_zit.csv");
	decl q = sizec(z);
	decl T = sizer(z);

	decl A_det_result = zeros(rep,1);
	decl d_result = zeros(rep,1);
	decl k_result = zeros(rep,1);

	decl keep_A_det_all = zeros(n_samples-burn);
	decl keep_d_all     = zeros(n_samples-burn);
	decl keep_k_all     = zeros(n_samples-burn);

	decl elapse = 0;

	decl elapse_5k = zeros(n_samples/5000 + 1, 1);
	
	for(decl hh = 0; hh < rep; ++hh)
	{
		decl time = timer(), j, t;	
		//Store Draws ===========================================
		decl keep_A_det = zeros(n_samples,1);
		decl keep_d     = zeros(n_samples,1);
		decl keep_k     = zeros(n_samples,1);
		//=======================================================

		decl t_5k = 0;
		
		//Set initial values ====================================
		decl gamma_0 = q;
		decl C0_inv = gamma_0 * unit(q);
		decl lambda0 = .2;
		decl A = unit(q);
		decl k = zeros(1,1); k = 15;
		decl d = zeros(1,1); d = 0.6;
 
		decl Qt = Initials(T, A, k, d);
		//========================================================

		// General settings for d and k
		decl ninit_d    = 3;
		decl ninit_k    = 5;	// This is for ARMS, as we do not know log-concavity, more initial support points are required
			     		// See Martino, Read, and Luengo (2012)
		decl dometrop_d = 1;
		decl dometrop_k = 1;
		decl nsamp = 1;

		//For d
		decl mdl = zeros(1,1); mdl[0][0]     = -0.99;
		decl mdr = zeros(1,1); mdr[0][0]     =  0.99;
		decl dprev = zeros(1,1); dprev[0][0] =  0.6;
		decl samp_d = zeros(1,nsamp); 	
		decl ansd = zeros(1,1);	
		d_post = new postd();	                    

		//For k							
		decl mkl = zeros(1,1); mkl[0][0]     = 4.1;
		decl mkr = zeros(1,1); mkr[0][0]     = 60;
		decl kprev = zeros(1,1); kprev[0][0] = 15;
		decl samp_k = zeros(1,nsamp);
		decl ansk = zeros(1,1);  
			 k_post = new postk();
  
		//Start the sampling!
		for(j=0; j<n_samples; ++j)
		{ 
			//Update Qt, t = 1, 2,..., T-1
			decl Qu, Qd, Qv, Qu_can, Qd_can, Qv_can;  
			decl Q, Qinv, Q_neg_sqrt, S, S_hat;
			decl canQt, canQtinv, Q_negsqrt_can, S_can, Qinv_star_can, P_can;
			decl Q_cur, Q_negsqrt_cur, Qu_cur, Qd_cur, Qv_cur;
			decl Qinv_star_cur, S_cur, P_cur;
			decl f_Qinv, f_Qinv_c;
   
			for(t=1; t<T; ++t)
			{
			    Q = Qt[t-1][][];
				decsvd(Q, &Qu, &Qd, &Qv);
				Q_neg_sqrt = Qu * diag(Qd.^(-d/2)) * Qv';
				S = (1/k) * Q_neg_sqrt * A * Q_neg_sqrt;		
				S_hat = invert(invert(S) + z[t-1][]' * z[t-1][]);
		
				//Use Self Writtern Wishart DGP and svd
				canQtinv = rwish(k+1, S_hat);                  //Use Wishart generator
				canQt    = invert(canQtinv);

				//Original definition  
				decsvd(canQt, &Qu_can, &Qd_can, &Qv_can);		   //For candidate.
				Q_negsqrt_can = Qu_can * diag(Qd_can.^(-d/2)) * Qv_can';
				S_can         = (1/k) * Q_negsqrt_can * A * Q_negsqrt_can; 	
				Qinv_star_can = diag(diagonal(canQt).^-.5);
				P_can         = Qinv_star_can * canQt * Qinv_star_can;

				f_Qinv        = sqrt(determinant(invert(P_can))) * ( determinant(canQtinv)^((-1-k*d)/2) ) *		 
								exp((-0.5) * trace(invert(S_can) * invert(Qt[t+1][][])) -
								0.5*trace( (invert(P_can)-canQtinv) * z[t-1][]' * z[t-1][]));

				Q_cur         = Qt[t][][];				 
				decsvd(Q_cur, &Qu_cur, &Qd_cur, &Qv_cur);
				Q_negsqrt_cur = Qu_cur * diag(Qd_cur.^(-d/2)) * Qv_cur';
				S_cur         = (1/k) * Q_negsqrt_cur * A * Q_negsqrt_cur;
				Qinv_star_cur = diag(diagonal(Q_cur).^-.5);			   
				P_cur         = Qinv_star_cur * Q_cur * Qinv_star_cur;

	  
				f_Qinv_c      = sqrt(determinant(invert(P_cur))) * ( determinant(invert(Q_cur))^((-1-k*d)/2) ) *	  
								exp((-0.5) * trace(invert(S_cur) * invert(Qt[t+1][][])) -
								0.5*trace((invert(P_cur)-invert(Q_cur)) * z[t-1][]' * z[t-1][]));	

						
				decl R = f_Qinv / f_Qinv_c;

				if(R > ranu(1,1))
				{     
					Qt[t][][] = canQt;
				}
			}

			//Update QT
			decl Qu_T, Qd_T, Qv_T;
			decl ST_can, ST_hat;
			decl canQT, canQT_inv, Q_negsqrt_can_T, Qinv_star_can_T, PT_can;
			decl QT_cur, Q_negsqrt_cur_T, Qinv_star_cur_T, PT_cur, f_QinvT, f_QinvT_c;
   
			ST_hat = invert(invert(S_cur) + z[T-1][]'*z[T-1][]);

			canQT_inv  = rwish(k+1, ST_hat); 
			canQT      = invert(canQT_inv);

			decsvd(canQT, &Qu_T, &Qd_T, &Qv_T);
			Q_negsqrt_can_T = Qu_T * diag(Qd_T.^(-d/2)) * Qv_T';
			ST_can          = (1/k) * Q_negsqrt_can_T * A * Q_negsqrt_can_T;
			Qinv_star_can_T = diag(diagonal(canQT).^-.5);
			PT_can          = Qinv_star_can_T * canQT * Qinv_star_can_T;

			f_QinvT         = (determinant(PT_can)^-.5) * (determinant(canQT_inv)^-.5) * 
							  exp( (-0.5)*trace( (invert(PT_can)-canQT_inv) * z[T-1][]' *  z[T-1][] ) );
						  
			//current
			QT_cur          = Qt[T][][];
			Qinv_star_cur_T = diag(diagonal(QT_cur).^-.5);
			PT_cur          = Qinv_star_cur_T * QT_cur * Qinv_star_cur_T;

		
			f_QinvT_c       = (determinant(PT_cur)^-.5)* (determinant(invert(QT_cur))^-.5) *
							  exp( (-0.5)*trace( (invert(PT_cur)-invert(QT_cur)) * z[T-1][]' * z[T-1][] ) );


			decl RT = f_QinvT / f_QinvT_c;
                       
			if(RT > ranu(1,1))
			{                                 
				Qt[T][][]  = canQT;
			}

			//Update A ==================================================================================   
			decl A_inv, Au, Ad, Av, Q_sqrt, Cinv_sum=zeros(q,q);
			decl prod_A = new array[T];
				 prod_A[0:T-1] = zeros(q,q);
		
			for(decl t=0; t<T; ++t)
			{
				decsvd(Qt[t][][], &Au, &Ad, &Av);
				Q_sqrt = Au * diag(Ad.^(d/2)) * Av';
				prod_A[t][][] = Q_sqrt * invert(Qt[t+1][][]) * Q_sqrt;
				Cinv_sum += prod_A[t][][];
			}
	 
			decl Cinv = k * Cinv_sum;

			A_inv = rwish((k*T + gamma_0), invert(C0_inv + Cinv));
			A     = invert(A_inv);
	
			keep_A_det[j] = determinant(A);	
			//===========================================================================================

			//Update d ==================================================================================
			decl sum_log_det = 0;			 // to compute psi
			for(decl t=1; t<=T; ++t)
			{
				sum_log_det += log(determinant(invert(Qt[t-1][][])));
			}
		
			decl psi = (-k/2) * sum_log_det;

			d_post -> set_other(psi, A_inv, Qt, k, T, q);
			
			ansd        = Arms_simple(post_d, ninit_d, mdl, mdr, dometrop_d, dprev, samp_d);
			dprev[0][0] = samp_d[0][0];
			d           = samp_d[0][0];
			keep_d[j]   = d;
			//===========================================================================================

			//Update k ==================================================================================  
			decl sum_log_Qt = 0;
			decl Cinvk = zeros(q,q);
			decl prod_k = new array[T];
				 prod_k[0:T-1] = zeros(q,q);
	  
			decl Q_sqrt_k, Qu_k, Qd_k, Qv_k;
			for(decl t=0; t<T; ++t)
			{
				decsvd(Qt[t][][], &Qu_k, &Qd_k, &Qv_k);
				Q_sqrt_k = Qu_k * diag(Qd_k.^(d/2)) * Qv_k';
 
				prod_k[t][][] = Q_sqrt_k * invert(Qt[t+1][][]) * Q_sqrt_k;
				sum_log_Qt   += log(determinant(prod_k[t][][]));
				Cinvk        += prod_k[t][][];
			}

			k_post -> set_other_k(lambda0, A, Cinvk, sum_log_Qt, T, q);
			ansk   = Arms_simple(post_k, ninit_k, mkl, mkr, dometrop_k, kprev, samp_k);

			kprev[0][0] = samp_k[0][0];
			k           = samp_k[0][0];
			keep_k[j]   = k;

			if(fmod(j+1, 5000) == 0)
			{
				t_5k = t_5k | (timer()-time)/100;
			}			
		}
		
		keep_A_det_all = keep_A_det_all ~ keep_A_det[burn:(n_samples-1)];
		keep_d_all     = keep_d_all ~ keep_d[burn:(n_samples-1)];
		keep_k_all     = keep_k_all ~ keep_k[burn:(n_samples-1)];

		elapse_5k      = elapse_5k ~ t_5k;		
		elapse = elapse | (timer()-time)/100;
		print("hh = ", hh, "\n");
	}

	savemat("C:/Research/ARMS_NoSMW_A_2.csv", keep_A_det_all);
	savemat("C:/Research/ARMS_NoSMW_d_2.csv", keep_d_all);
	savemat("C:/Research/ARMS_NoSMW_k_2.csv", keep_k_all);
	savemat("C:/Research/ARMS_NoSMW_time_2.csv", elapse);	
	savemat("C:/Research/ARMS_NoSMW_time_5k_2.csv", elapse_5k);
	
	print("This is DHKT ARMS + NoSMW formula.", "\n");
	return 0;
}
